/* Multipattern random searches using set Horspool's algorithm adapted to 
search inside a text compressed with End-Tagged Dense Code.
Lightweight Natural Language Text Compression: Information Retrieval 2006
Copyright (C) 2005 Antonio Faria.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Author's contact: Antonio Faria, Dept. of Computer Science, University of
A Corua. Campus de Elvia s/n. Spain  fari@udc.es
*/

/*-----------------------------------------------------------------------
  search.c:     Implements a trie and deals with the search of a set of patterns
  inside a compressed file (searchFullFile()). Searching the whole file is done 
  one buffer at a time.
 ------------------------------------------------------------------------*/
 
#include "searcher.h"


/***********************************************************************************************/
/*------------------------------------------------------------------
   Given a codeword stored as a number, it 
   obtains the number of bytes and writes they to *code as an array.	
------------------------------------------------------------------*/
void getBytes(unsigned char *code, unsigned long number, int numBytes) {
	unsigned long tmp;
	int i;
	char *pt;
	
	tmp = LE_LONG32(number);
	tmp = LINE_UP (tmp,numBytes);
	pt = (unsigned char *)&tmp;
	
	for (i=0;i<numBytes;i++) {
		code[i]= pt[i];				
	}	
}

/*------------------------------------------------------------------
 GetCode
 gets the codeword of a word ranked "position"
 returns an unsigned long value (up to 4 bytes codewords) in "*codeword"
         the number of bytes of the codeword.
------------------------------------------------------------------*/
void getCode(register unsigned long position,
						register unsigned long *codeword, register unsigned long *k){
	unsigned long despl;

	*k=1;
	*codeword = (position& 0x7F);
	position  >>= 7;
	despl=8;

	while (position > 0) {
		position -=1;
		*codeword |= (( (position & 0x7F) | 0x80 )<<despl );
		despl += 8;
		position  >>= 7;
		(*k) ++;
	}
}

/*------------------------------------------------------------------
 Prints a code: debuging purposes only
------------------------------------------------------------------*/
void printCodigo (byte *cw, int size) {
	int i;
	byte *c;
	c=cw;
	fprintf(stderr,"\nINICIO PRINTCODIGO codeSize %d: CODIGO:",size);
	for (i=0;i<size;i++) {fprintf(stderr,"[%d]",*(c++));}
}


/***********************************************************************************************/

/*------------------------------------------------------------------
 Initialize the list of search patterns from the input parameters
 ------------------------------------------------------------------*/
 t_patterns initializePatterns(int numpatts, byte **patts, int *lens) {
 	t_patterns mm;
 	unsigned int p, initP;
 	unsigned int m_p;
 	byte *ch;

 	mm = (t_patterns) malloc (sizeof(struct t_pats));
 	mm->lmax =  MAX_CODEWORD_SIZE;
 	mm->lmin = INFINITE;

 	mm->size = numpatts;

 	mm->patterns = (t_pattern *) malloc (sizeof(t_pattern) * mm->size);

	for (p =0; p < mm->size;p++) {
 		m_p =lens[p];
		mm->patterns[p].size = m_p;

 		ch = mm->patterns[p].word;
 		strncpy(ch, patts[p], m_p);

		mm->patterns[p].matches = 0;
		mm->patterns[p].flag = p;

		mm->lmin = MIN(mm->lmin,m_p);
		mm->lmax = MAX(mm->lmax,m_p);
	}
 
	//	fprintf(stderr,"\nparameters [%d]",mm->size);
	//	fprintf(stderr,"\n lmin [%d]",mm->lmin);
	//	fprintf(stderr,"\n lmax[%d]",mm->lmax);
	 	
 	return mm;
}


/*******************************************************************************/
/*****     OPERATIONS IN THE HORSPOOL's D-VECTOR    ****************************/
/*******************************************************************************/
/*------------------------------------------------------------------
  Initializes Horspool's d-vector (jumps for each letter)
 ------------------------------------------------------------------*/
void inicialize_d(t_patterns P, byte *d) {
	unsigned int c,i,m_i,k;

	for (c=0;c<256;c++) {d[c]=P->lmin;}

	for (i=0;i<P->size;i++) {

		m_i= P->patterns[i].size;
		for (k=0; k<m_i-1; k++) {
			c=P->patterns[i].word[k];
			d[ c ] = MIN ( d[c], m_i - k-1);
		}
	}
}


/*******************************************************************************/
/*****     OPERATIONS WITH THE TRIE         ************************************/
/*******************************************************************************/

/*------------------------------------------------------------------
 Initializes the trie used by the searcher, from the list of patterns
 ------------------------------------------------------------------*/
t_trie inicializeTrie(t_patterns P) {
	register unsigned int s,ch;
	register int j,i,m_i;
	register byte *c;
	t_trie trie;

	trie = (t_trie) malloc (sizeof(struct t_ttrie));

	//initialize trie to all zeroes:
	//REVISAR, ES POSIBLE QUE NO HAGA FALTA SI INICIALIZO DE CADA VEZ.
	for (s=0;s<MAX_STATES;s++) {
		trie->nodes[s].n=0;
		trie->nodes[s].f=NORMAL;

		for (ch=0;ch<256;ch++) {
			trie->transitions[s][ch]=ROOT_STATE;
		}
	}

	trie->zeroNode = 0;
	//initializes stack with all the free states.
	trie->stack = createStack(MAX_STATES);
	for (i=MAX_STATES-1;i>0;i--) {
		push(trie->stack,i);
	}
	
	//adds patterns to the trie
	for (i=0; i<P->size; i++) {  //for all patterns i
		addToTrie(trie, P->patterns[i].word, P->patterns[i].size, P->patterns[i].flag);
	}

	return trie;
}


/*------------------------------------------------------------------
 Debuging only
 Shows the states and transitions of the trie
 ------------------------------------------------------------------*/
void printTrie(t_trie trie,int maxstate) 	{
	int z,k;

	fprintf(stderr,"\n\n DATA IN THE TRIE \n*******************");
	fprintf(stderr,"\n zeronode [%d] ", trie->zeroNode);
	fprintf(stderr,"\n STATES... ");
	for (z =0; z< maxstate; z++) {
		fprintf(stderr,"\n state[%d]:: n=%d :: flag=%d ",z,trie->nodes[z].n,trie->nodes[z].f);
		for (k=0;k<256;k++) {
			if (trie->transitions[z][k] !=ROOT_STATE) {
				fprintf(stderr,"(%d ---> %d).. ",k,trie->transitions[z][k]);
			}
		}
	}
}

/*------------------------------------------------------------------
 Adds a new pattern to the trie (states and transitions)
 ------------------------------------------------------------------ */
void addToTrie(t_trie trie, byte *nuevo, int size, unsigned int flag) {

	register byte *c;
	unsigned int state,newState;

	//while transitions are possible with the current pattern...

	state = ROOT_STATE;
	c = nuevo+size -1;  //pointer TO last char of the word

	while ((c>=nuevo) AND (trie->transitions[state][*c] !=ROOT_STATE)) {
		trie->nodes[state].n ++;
		state = trie->transitions[state][(*c)];
		c--;  //pattern from right to left in the trie.
	}

	//add new states to the trie:     at this moment,...
	if (c<nuevo) {
	trie->nodes[state].f = flag;
	trie->nodes[state].n++;
	}
	else {
		while (c>=nuevo) {
			newState = pop(trie->stack);
			trie->transitions[state][*c]= newState; //transition for current state,
							        // with character *c
			trie->nodes[state].n++; // OR also --> trie[s][*c].n ++;
			trie->nodes[newState].f=NORMAL;
			state = newState;
			c--;
		}
		//changes newState flag
		trie->nodes[newState].f = flag;
		trie->nodes[newState].n++;
	}
}

/*******************************************************************************/
/*******************************************************************************/
/*****     SEARCH PROCESS INSIDE  OF TEXT     **********************************/
/*******************************************************************************/
/*******************************************************************************/

/*------------------------------------------------------------------
 Performs search process inside a Buffer of text.
 ------------------------------------------------------------------ */
void searchInBufferETDC (BufferRead B, byte **begin, byte **end,unsigned long pos,
					 t_patterns P, t_trie trie, byte *d ){

	register unsigned int state;
	register unsigned int flag;

	register byte *cend;
	register byte *cpos;
	register byte *cj;

	cpos=*begin + pos;
	cend=*end;

	while ( 1 )   {
		state = ROOT_STATE;
		
		/*** skip loop  ***/
		while ( ! trie->transitions[state][*cpos] ) {
				#ifdef CONTADOR_BYTES_ENABLED    
			LEIDOS += d[*cpos];			
				#endif
		cpos += d[*cpos];
		}
		
		/*** normal loop  ***/
		cj= cpos;

		while (trie->transitions[state][*cj])  { //	while ((trie->transitions[state][*cj] !=ROOT_STATE) ) 
													//while	( ( cj>=cbegin ) AND (trie->transitions[state][*cj] !=ROOT_STATE) ) {

			state = trie->transitions[state][*cj];
			flag  = trie->nodes[state].f;

			if (flag != NORMAL) {
				//matching a break-code ??
				if (cpos >=cend) { // if (flag == BREAKCODE)
				
					cj=cpos -1;
							#ifdef CONTADOR_BYTES_ENABLED
						LEIDOS --;
							#endif
					while (*(cj) >=128) {
						cj--;
							#ifdef CONTADOR_BYTES_ENABLED
						LEIDOS --;
							#endif
							}
					cpos = cj;
					cpos += d[*cj]; //calculates next jump before exitting
											
					*begin = cpos - B->minDataInBuffer;								 
					return;					 
				}	
									
				//a true match ?
				{
					cj--;
					if ((*cj) <128) {
						P->patterns[flag].matches ++;			
						break;
					}
					cj++;
				}
			}//flag != normal
			cj--;
		}//inner while
			
			#ifdef CONTADOR_BYTES_ENABLED
		LEIDOS += d[*cpos];	
			#endif
		cpos += d[*cpos];
			
	} //while ( 1 )

	*begin = cpos - B->minDataInBuffer;
}


/*------------------------------------------------------------------
 * Performs search process over the whole file.
------------------------------------------------------------------ */
void searchFullFile (char *infile, int numWords, byte **words, int *lens){
	BufferRead B;
	t_trie trie;
	byte d[256];
	int f;
	
	byte AUX[4];  	//Used to save the 1-4 bytes that are modified in the buffer 
					//to put a break code at the end of the buffer.

	byte *text,*top, *nextbuf;  //pointers to the input buffer

	t_patterns P;

		#ifdef CONTADOR_BYTES_ENABLED
	LEIDOS = 0;
		#endif


	//initialization of the search patterns 
	P= initializePatterns(numWords, words,lens);

	//initializes Horspool jumps vector "d[]" (used to jump forwards through the file)
	inicialize_d(P, d);

	//initilizes the trie.
	trie = inicializeTrie(P);
		//printTrie(trie,20);

	// Initialization of the structure Buffer used to read the file
	f = open (infile, O_RDONLY);
	if (f == -1) {
		fprintf(stderr,"Cannot read file %s",*infile);
		exit(0);
	}

	B= bufCreate(P->lmax);
	bufSetFile (B,f);

	// Search process. ................

	fprintf(stderr, "\n STARTING SEARCH......");
	{ unsigned long i;
		byte *NEWEND;

		byte *pbeg,*pend,*codeStart;
		unsigned long pos;

		pos=P->lmin-1;

			#ifdef CONTADOR_BYTES_ENABLED
		LEIDOS = pos;
			#endif

		//process the whole file by reading one buffer of data at a time and then performing
		//set-horspool (modified) over those buffers of data.
		while (!bufEmpty(B) ) {

			// set the area of the buffer that will be processed.
			bufCurrent (B,(byte**)&text,(byte**)&top);
			pbeg = text; 
			pend = top;

			// copies a searched pattern (the first one) to the end of the buffer as a BREAK-CODE
			// in order to avoid the need of  checking for the "end of buffer" in the search process.
			// Since some bytes at the end of the buffer are replaced, the are saved in AUX first.
			NEWEND = pend+1;
			for (i=0;i<lens[0];i++){ 
				AUX[i]=NEWEND[i];
				NEWEND[i]=words[0][i]; 
			}
					
			// Searching inside the buffer 
			#ifdef VERSIONETDC
				searchInBufferETDC (B, &pbeg, &pend,pos, P, trie, d );
			#endif

			//the bytes at the end of the buffer are restored to the buffer from AUX.
			NEWEND = pend+1;
			for (i=0;i<lens[0];i++){ 
				NEWEND[i]=AUX[i];					
			}
			
			// Reads the next buffer of data that will be processed in the next iteration
			pos = B->dsize > B->minDataInBuffer? B->minDataInBuffer: B->dsize;
			if (bufEof(B)) nextbuf=top;
			else nextbuf=pbeg;

			bufLoad (B, (char*)nextbuf);
		}//while
	}

	fprintf(stderr, "... search FINISHED!!\n");

	// Summarizes the results obtained.
	// If login is enabled, it writes, for each pattern, the number of occurrencias found.
	
	{//counts matches
		long counter;
		counter = 0;
		for (f=0;f<P->size;f++) { 
			counter += P->patterns[f].matches;
		}	
		fprintf(stderr, "   Accumulated matches found: %ld \n",counter);
	}
	
	#ifdef LOGGING_ENABLED     
		FILE *fLOG2;
	
		if ((fLOG2 = fopen("log2","w")) == NULL)
			{
			  fprintf(stderr,
				  "Error en la apertura del fichero de tiempos %s\n",
				  "tiempos");
			  exit(0);
			}

		fprintf(stderr,"\n SUMMARY OF MATCHES FOUND: LMIN %d, number of patterns %d\n",P->lmin,P->size);
		for (f=0;f<P->size;f++) {
			fprintf(stderr," %ld.- w[%s], s[%d], matches %5d \n",f,P->patterns[f].word,P->patterns[f].size,P->patterns[f].matches);
			fprintf(fLOG2,"%5d.- %6d\n",f,P->patterns[f].matches);
		}
			
		fflush(fLOG2);
		fclose(fLOG2);
	#endif 	

	#ifdef CONTADOR_BYTES_ENABLED    
		fprintf(stderr,"\n TOTAL BYTES READ (or jumped): %ld \n",	LEIDOS);
	#endif

	//Frees memory...........
	bufDestroy(B);

	free(P->patterns);
	free(P);

	destroyStack (trie->stack);
	free(trie);

} //searchFullFile


